/* ------------------------------------------------------------------------- */
/*
 *  argument.h
 *
 *  Copyright (c) 2004 - 2008 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Fri 08 Feb 2008 04:09:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_ARGUMENT_H
#define CLX_ARGUMENT_H

#include <string>
#include <vector>
#include <map>
#include "lexical_cast.h"
#include "tokenizer.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	/*!
	 *  basic_argument_container
	 *
	 *  BNF description of the argument is as follow:
	 *  <program> [<nokey_param>]* (['-'<key>] | ['-'+<key> [<param>]*])*
	 *
	 *  The basic_argument_container class has two containers. One is
	 *  std::map<key, param>, and the other is std::vector<nokey_param>.
	 *  This class provides the interface that the upper class can control
	 *  these two containers as one container.
	 */
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_argument_container {
	public:
		typedef CharT char_type;
		typedef unsigned int size_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef std::map<string_type, std::vector<string_type> > container;
		typedef typename container::key_type key_type;
		typedef typename container::mapped_type mapped_type;
		typedef typename container::value_type value_type;
		
		// constructor and destructor
		basic_argument_container() : v_(), head_() {}
		
		virtual ~basic_argument_container() {}
		
		// operator
		bool insert(const key_type& x, const string_type& s) {
			typedef typename container::iterator iterator;
			
			// if the argument is <nokey_param>, x is empty.
			if (x.empty()) {
				head_.push_back(s);
				return true;
			}
			
			iterator pos = v_.find(x);
			if (pos == v_.end()) {
				mapped_type v;
				v.push_back(s);
				std::pair<iterator, bool> r = v_.insert(value_type(x, v));
				return r.second;
			}
			if (!s.empty()) pos->second.push_back(s);
			return true;
		}
		
		// access
		mapped_type operator[](const key_type& x) {
			if (x.empty()) return head_;
			else return v_[x];
		}
		
		bool exist(const key_type& x) {
			if (v_.find(x) != v_.end()) return true;
			else return false;
		}
		
	private:
		container v_;
		mapped_type head_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_argument
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_argument {
	public:
		typedef CharT char_type;
		typedef int int_type;
		typedef unsigned int size_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef basic_argument_container<CharT, Traits> container;
		typedef typename container::mapped_type param_list;
		
		// constructor and destructor
		explicit basic_argument(char_type c = '-') :
			identifier_(c), program_(), v_() {}
		
		explicit basic_argument(int_type argc, char_type* argv[], char_type c = '-') :
			identifier_(c), program_(), v_() {
			this->scan(argc, argv);
		}
		
		virtual ~basic_argument() {}
		
		// operator
		basic_argument& assign(int_type argc, char_type* argv[]) {
			program_ = argv[0];
			string_type key;
			bool eoh = false; // End of head parameter
			for (int_type i = 1; i < argc; i++) {
				string_type tmp(argv[i]);
				if (tmp.at(0) == identifier_) {
					eoh = true;
					if (tmp.at(1) != identifier_ && tmp.size() > 2) {
						for (size_type j = 1; j < tmp.size(); j++) {
							key = tmp.at(j);
							v_.insert(key, "");
						}
						key = "";
					}
					else {
						key = tmp.substr(tmp.find_first_not_of(identifier_));
						if (i + 1 == argc || argv[i+1][0] == identifier_) v_.insert(key, "");
					}
				}
				else {
					if (eoh && key.empty()) break;
					v_.insert(key, tmp);
				}
			}
			return *this;
		}
		
		// access
		string_type program() { return program_; }
		param_list head() { return v_[string_type("")]; }
		param_list operator[](const string_type& x) { return v_[x]; }
		
		param_list operator[](const char_type* x) {
			string_type tmp(x);
			return (*this)[tmp];
		}
		
		param_list operator[](char_type x) {
			string_type tmp(1, x);
			return (*this)[tmp];
		}
		
		bool exist(const string_type& x) {
			functor f(x);
			for (size_type i = 0; i < f.size(); i++) {
				if (v_.exist(f.at(i))) return true;
			}
			return false;
		}
		
		bool exist(const char_type* x) {
			string_type tmp(x);
			return this->exist(tmp);
		}
		
		bool exist(char_type x) {
			string_type tmp(1, x);
			return v_.exist(tmp);
		}
		
		template <class ValueT>
		basic_argument& operator()(ValueT& dest, size_type n = 0) {
			return this->get_param(dest, n);
		}
		
		template <class ValueT>
		basic_argument& operator()(const string_type& key, ValueT& dest, size_type n = 0) {
			return this->get_param(key, dest, n);
		}
		
		template <class ValueT>
		basic_argument& operator()(const char_type* key, ValueT& dest, size_type n = 0) {
			string_type tmp(key);
			return (*this)(tmp, dest, n);
		}
		
		template <class ValueT>
		basic_argument& operator()(char_type key, ValueT& dest, size_type n = 0) {
			string_type tmp(1, key);
			return (*this)(tmp, dest, n);
		}
		
	private:
		typedef basic_tokenizer<escape_separator<CharT, Traits>, string_type> functor;
		
		char_type identifier_;
		string_type program_;
		container v_;
		
		template <class ValueT>
		basic_argument& get_param(ValueT& dest, size_type n) {
			if (n < this->head().size()) dest = lexical_cast<ValueT>(this->head().at(n));
			return *this;
		}
		
		template <class ValueT>
		basic_argument& get_param(const string_type& key, ValueT& dest, size_type n) {
			functor f(key);
			for (size_type i = 0; i < f.size(); i++) {
				if (v_.exist(f.at(i)) && n < v_[f.at(i)].size() && !v_[f.at(i)].at(0).empty()) {
					dest = lexical_cast<ValueT>(v_[f.at(i)][n]);
				}
			}
			return *this;
		}
	};
	
	typedef basic_argument<char> argument;
}

#endif // CLX_ARGUMENT_H
