/* ------------------------------------------------------------------------- */
/*
 *  format.h
 *
 *  Copyright (c) 2004 - 2007 "clown", All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Sat 10 Mar 2007 20:58:00 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_FORMAT_H
#define CLX_FORMAT_H

#include <cstdlib>
#include <cctype>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <ostream>
#include <string>
#include "lexical_cast.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_format
	/* --------------------------------------------------------------------- */
	template <
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_format {
	public:
		typedef CharT char_type;
		typedef unsigned int size_type;
		typedef typename std::basic_string<CharT, Traits> string_type;
		
		// constructor and destructor
		basic_format() : str_(), fmt_() {}
		
		explicit basic_format(const char_type* fmt) :
			str_(), fmt_(fmt) { this->reset(); }
		
		explicit basic_format(const string_type& fmt) :
			str_(), fmt_(fmt) { this->reset(); }
		
		// operator
		template <class ValueT>
		basic_format& operator%(const ValueT& x) {
			if (*pos_ == rep && pos_ + 1 != fmt_.end() && *(pos_ + 1) != rep) {
				string_type flag;
				while (++pos_ != fmt_.end() && types_.find(*pos_) == string_type::npos) flag += *pos_;
				if (pos_ == fmt_.end()) return *this; // uncorrect format
				str_ += this->put(x, *pos_++, flag);
			}
			while (pos_ != fmt_.end() && *pos_ != rep) str_ += *pos_++;
			return *this;
		}
		
		template <class Ch, class Tr>
		friend std::basic_ostream<Ch, Tr>& operator<<(
			std::basic_ostream<Ch, Tr>& sout, const basic_format<Ch, Tr>& f) {
			sout << f.str_;
			return sout;
		}
		
	private:
		typedef typename string_type::const_iterator const_iterator;
		typedef typename std::basic_stringstream<CharT, Traits> internal_stream;
		
		static const char_type rep = '%';
		static const char_type lpos = '-';
		static const char_type ws = ' ';
		static const char_type sign = '+';
		static const char_type zero = '0';
		static const char_type prefix = '#';
		static const char_type dot = '.';
		
		string_type types_;
		string_type str_;
		string_type fmt_;
		const_iterator pos_;
		
		void reset() {
			str_.clear();
			pos_ = fmt_.begin();
			types_ = "diouxXeEfFgGcs";
			while (pos_ != fmt_.end() && *pos_ != rep) str_ += *pos_++;
		}
		
		template <class ValueT>
		string_type put(const ValueT& x, char_type c, const string_type& flag) {
			internal_stream ss;
			
			this->setopt(ss, flag);
			this->settype(ss, c);
			ss << x;
			
			return ss.str();
		}
		
		void setopt(internal_stream& ss, const string_type& flag) {
			bool lp = false;
			for (const_iterator pos = flag.begin(); pos != flag.end(); pos++) {
				switch (*pos) {
				case lpos:
					lp = true;
					ss << std::setfill(ws);
					ss << std::left;
					break;
				case prefix:
					ss << std::showbase;
					break;
				case sign:
					ss << std::showpos;
					break;
				case ws:
					ss << std::setfill(*pos);
					break;
				case zero:
					if (!lp) ss << std::setfill(*pos);
					break;
				case dot:
					this->setprecision(ss, ++pos, flag.end());
					if (pos == flag.end()) return;
					break;
				default:
					if (isdigit(*pos)) this->setwidth(ss, pos, flag.end());
					if (pos == flag.end()) return;
					break; /* ignore unknown flags */
				}
			}
		}
		
		void setwidth(internal_stream& ss, const_iterator& cur, const_iterator last) {
			std::string digit;
			for (; cur != last && isdigit(*cur); cur++) digit += *cur;
			if (!digit.empty()) ss << std::setw(lexical_cast<int>(digit));
			if (cur != last && *cur == dot) this->setprecision(ss, ++cur, last);
		}
		
		void setprecision(internal_stream& ss, const_iterator& cur, const_iterator last) {
			std::string digit;
			for (; cur != last && isdigit(*cur); cur++) digit += *cur;
			if (!digit.empty()) ss << std::setprecision(lexical_cast<int>(digit));
		}
		
		void settype(internal_stream& ss, char_type c) {
			char_type ctype;
			if (isupper(c)) ss << std::uppercase;
			ctype = tolower(c);
			
			switch (ctype) {
			case 'o':
				ss << std::oct;
				break;
			case 'x':
				ss << std::hex;
				break;
			case 'e':
				ss << std::scientific;
				break;
			case 'f':
				ss << std::setiosflags(std::ios::fixed);
				break;
			default:
				break;
			}
		}
	};
	
	typedef basic_format<char> format;
}
#endif // CLX_FORMAT_H
