/* ------------------------------------------------------------------------- */
/*
 *  sockaddress.h
 *
 *  Copyright (c) 2004 - 2008 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Mon 05 Feb 2008 15:26:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_SOCKADDRESS_H
#define CLX_SOCKADDRESS_H

#include <cstdlib>
#include <cstring>
#include <string>
#include <sstream>
#include <exception>
#include "socket.h"
#include "resolver.h"

namespace clx {
	/* --------------------------------------------------------------------- */
	//  sockaddress_error
	/* --------------------------------------------------------------------- */
	class sockaddress_error : public std::exception {
	public:
		typedef char char_type;
		typedef std::string string_type;
		typedef std::stringstream sstream_type;
		
		sockaddress_error() : std::exception(), hostname_() {}
		
		explicit sockaddress_error(const string_type& host) :
			std::exception(), hostname_(host) {}
		
		explicit sockaddress_error(const char_type* host) :
			std::exception(), hostname_(host) {}
		
		virtual ~sockaddress_error() throw() {}
		
		virtual const char_type* what() {
			sstream_type tmp;
			tmp << this->hostname() << ": Non-existent address";
			return tmp.str().c_str();
		}
		
		const string_type& hostname() const { return hostname_; }
	
	private:
		string_type hostname_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_sockaddress
	/* --------------------------------------------------------------------- */
	template <
		int Family,
		class CharT = char,
		class Traits = std::char_traits<CharT>
	>
	class basic_sockaddress {
	public:
		typedef basic_resolver<Family> resolver;
		typedef struct sockaddr_in inet_type;
		typedef struct sockaddr base_type;
		typedef struct in_addr address_type;
		typedef CharT char_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef unsigned int size_type;
		
		// constructor and destructor
		basic_sockaddress() : addr_() {}
		
		explicit basic_sockaddress(const inet_type& addr) : addr_(addr) {}
		
		explicit basic_sockaddress(int port) : addr_() {
			this->assign(port);
		}
		
		explicit basic_sockaddress(const char_type* host, int port) : addr_() {
			this->assign(host, port);
		}
		
		explicit basic_sockaddress(const string_type& host, int port) : addr_() {
			this->assign(host, port);
		}
		
		virtual ~basic_sockaddress() {}
		
		basic_sockaddress& operator=(const inet_type& addr) {
			return this->assign(addr);
		}
		
		// operator
		basic_sockaddress& assign(const inet_type& addr) {
			addr_ = addr;
			return *this;
		}
		
		basic_sockaddress& assign(int port) {
			std::memset((char*)&addr_, 0, sizeof(addr_));
			addr_.sin_family = Family;
			addr_.sin_port = htons(port);
#ifdef CLX_WIN32
			addr_.sin_addr.S_un.S_addr = ::htonl(INADDR_ANY);
#else
			addr_.sin_addr.s_addr = ::htonl(INADDR_ANY);
#endif // CLX_WIN32
			return *this;
		}
		
		basic_sockaddress& assign(const char_type* host, int port, int n = 0) {
			address_type tmp;
			if (!::inet_aton(host, &tmp)) {
				resolver inet;
				inet.lookup(host);
				if (inet.addresses().empty()) throw sockaddress_error(host);
				tmp = inet.addresses().at(n);
			}
			
			std::memset(&addr_, 0, sizeof(addr_));
			addr_.sin_family = Family;
			addr_.sin_port = htons(port);
			std::memcpy((char*)&addr_.sin_addr, &tmp, sizeof(tmp));
			return *this;
		}
		
		basic_sockaddress& assign(const string_type& host, int port, int n = 0) {
			return this->assign(host.c_str(), port, n);
		}
		
		// get information about socket address
		const address_type& address() const { return addr_.sin_addr; }
		int port() const { return ::ntohs(addr_.sin_port); }
		
		size_type size() const { return sizeof(addr_); }
		const inet_type& reference() const { return addr_; }
		const inet_type* data() const { return &addr_; }
		const base_type* base_data() const { return (base_type*)this->data(); }
		
	private:
		inet_type addr_;
	};
}

#endif // CLX_SOCKADDRESS_H
