/* ------------------------------------------------------------------------- */
/*
 *  socket.h
 *
 *  Copyright (c) 2004 - 2007 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Fri 16 Mar 2007 17:52:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_SOCKET_H
#define CLX_SOCKET_H

#ifndef CLX_WIN32
#if defined(_WIN32) || defined(__WIN32__) || defined(WIN32)
#define CLX_WIN32
#endif
#endif // CLX_WIN32

#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <string>
#include <sstream>
#include <stdexcept>

#ifdef CLX_WIN32
#include "winsock_util.h"
#include <windows.h>
#include <io.h>
#else
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#endif // CLX_WIN32

namespace clx {
#ifdef CLX_WIN32
	typedef SOCKET socket_type;
	typedef int socklen_type;
	typedef unsigned long sockint_type;
#else
	typedef int socket_type;
	typedef socklen_t socklen_type;
	typedef u_int32_t sockint_type;
#endif // CLX_WIN32
	
	/* --------------------------------------------------------------------- */
	//  socket_error
	/* --------------------------------------------------------------------- */
	class socket_error : public std::runtime_error {
	public:
		typedef char char_type;
		typedef std::string string_type;
		typedef std::stringstream sstream_type;
		
		static const int nbuf = 1024;
		
		explicit socket_error(const string_type& what_arg) :
			std::runtime_error(what_arg), errno_(errno) {}
		
		virtual const char_type* what() {
			sstream_type tmp;
			char message[nbuf];
#ifdef CLX_WIN32
			strerror_s(message, sizeof(message), errno_);
#else
			strerror_r(errno_, message, sizeof(message));
#endif
			tmp << std::runtime_error::what() << ": " << message;
			return tmp.str().c_str();
		}
		
	private:
		int errno_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_rawsocket
	/* --------------------------------------------------------------------- */
	template <
		int Type,
		int Family
	>
	class basic_rawsocket {
	public:
		// constructor and destructor
		basic_rawsocket() : s_(-1) {}
		
		explicit basic_rawsocket(const basic_rawsocket& cp) : s_(-1) {
			this->assign(cp);
		}
		
		explicit basic_rawsocket(const socket_type& s) : s_(-1) {
			this->assign(s);
		}
		
		virtual ~basic_rawsocket() { this->close(); }
		
		basic_rawsocket& operator=(const basic_rawsocket& cp) {
			return this->assign(cp);
		}
		
		basic_rawsocket& operator=(const socket_type& s) {
			return this->assign(s);
		}
		
		// operator
		basic_rawsocket& assign(const basic_rawsocket& cp) {
			if (&cp == this) return *this;
			this->close();
#ifdef CLX_WIN32
			if (cp.s_ != (socket_type)-1) {
				::DuplicateHandle(::GetCurrentProcess(), (HANDLE)cp.s_,
					::GetCurrentProcess(), (LPHANDLE)&s_,
					0, TRUE, DUPLICATE_SAME_ACCESS);
			}
#else
			if (cp.s_ != -1) s_ = ::dup(cp.s_);
#endif
			return *this;
		}
		
		basic_rawsocket& assign(const socket_type& s) {
			this->close();
			s_ = s;
			return *this;
		}
		
		socket_type open() {
			s_ = ::socket(Family, Type, 0);
			return s_;
		}
		
		void close() {
			if (s_ >= 0) {
#ifdef CLX_WIN32
				::closesocket(s_);
#else
				::close(s_);
#endif
				s_ = -1;
			}
		}
		
		socket_type socket() const { return s_; }
		
	private:
		socket_type s_;
	};
}

#endif // CLX_SOCKET_H
