/* ------------------------------------------------------------------------- */
/*
 *  sockstream.h
 *
 *  Copyright (c) 2004 - 2008 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Mon 05 Feb 2008 15:27:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_SOCKSTREAM_H
#define CLX_SOCKSTREAM_H

#include <iostream>
#include <streambuf>
#include <string>
#include <vector>

namespace clx {
	/* --------------------------------------------------------------------- */
	//  basic_sockbuf
	/* --------------------------------------------------------------------- */
	template <
		class Socket,
		class CharT,
		class Traits = std::char_traits<CharT>
	>
	class basic_sockbuf : public std::basic_streambuf<CharT, Traits> {
	public:
		typedef Socket descriptor;
		typedef typename std::vector<CharT> container;
		typedef typename container::size_type size_type;
		typedef CharT char_type;
		typedef Traits traits;
		typedef typename Traits::int_type int_type;
		
		// constructor and destructor
		explicit basic_sockbuf(const descriptor& s, size_type n) :
			std::basic_streambuf<CharT, Traits>(), sock_(s), buffer_(n) {}
		
		virtual ~basic_sockbuf() {}
		
		// operator
		bool is_open() const { return sock_.socket() >= 0; }
		
	protected:
		virtual int_type overflow(int_type c = traits::eof()) {
			if (!traits::not_eof(c)) return traits::eof();
			char tmp = static_cast<char>(c);
			if (sock_.send(&tmp, 1) <= 0) return traits::eof();
			return c;
		}
		
		virtual int_type underflow() {
			if (!this->is_open()) return traits::eof();
			if (this->gptr() >= this->egptr()) {
				if (this->gptr() > this->eback()) buffer_[0] = *(this->gptr() - 1);
				
				int l = sock_.recv(&buffer_[1], static_cast<int>(buffer_.size() - 1));
				if (l <= 0) return traits::eof();
				
				this->setg(&buffer_[0], &buffer_[1], &buffer_[l + 1]);
			}
			return *this->gptr();
		}
	
	private:
		descriptor sock_;
		container buffer_;
	};
	
	/* --------------------------------------------------------------------- */
	//  basic_sockstream
	/* --------------------------------------------------------------------- */
	template <
		class Socket,
		class CharT = char,
		class Traits = std::char_traits<CharT>
	>
	class basic_sockstream : public std::basic_iostream<CharT, Traits> {
	public:
		typedef Socket descriptor;
		typedef CharT char_type;
		typedef std::basic_string<CharT, Traits> string_type;
		typedef basic_sockbuf<Socket, CharT, Traits> sockbuf;
		typedef typename sockbuf::size_type size_type;
		
		static const size_type nbuf = 65536;
		
		// constructor and destructor
		explicit basic_sockstream(const descriptor& s, size_type n = nbuf) :
			std::basic_iostream<CharT, Traits>(0), sbuf_(s, n) {
			this->rdbuf(&sbuf_);
			if (!sbuf_.is_open()) this->setstate(std::ios::badbit);
		}
		
		explicit basic_sockstream(const char_type* host, int port, size_type n = nbuf) :
			std::basic_iostream<CharT, Traits>(0), sbuf_(descriptor(host, port), n) {
			this->rdbuf(&sbuf_);
			if (!sbuf_.is_open()) this->setstate(std::ios::badbit);
		}
		
		explicit basic_sockstream(const string_type& host, int port, size_type n = nbuf) :
			std::basic_iostream<CharT, Traits>(0), sbuf_(descriptor(host, port), n) {
			this->rdbuf(&sbuf_);
			if (!sbuf_.is_open()) this->setstate(std::ios::badbit);
		}
		
		virtual ~basic_sockstream() { this->rdbuf(0); }
		
		// operator
		bool is_open() const { return this->rdbuf()->is_open(); }
	
	private:
		sockbuf sbuf_;
	};
}

#endif // CLX_SOCKSTREAM_H
