/* ------------------------------------------------------------------------- */
/*
 *  tcp.h
 *
 *  Copyright (c) 2004 - 2008 "clown". All rights reserved.
 *
 *  Permission to copy, use, modify and distribute this software is
 *  granted provided this copyright notice appears in all copies.
 *  This software is provided "as is" without express or implied
 *  warranty, and with no claim as to its suitability for any purpose.
 *
 *  Last-modified: Sun 03 Feb 2008 22:31:03 JST
 */
/* ------------------------------------------------------------------------- */
#ifndef CLX_TCP_H
#define CLX_TCP_H

#include <string>
#include <sstream>
#include "socket.h"
#include "sockaddress.h"
#include "sockstream.h"

namespace clx {
	namespace tcp {
		/* ----------------------------------------------------------------- */
		//  basic_socket
		/* ----------------------------------------------------------------- */
		template <
			int Family,
			class CharT = char,
			class Traits = std::char_traits<CharT>
		>
		class basic_socket : public basic_rawsocket<SOCK_STREAM, Family> {
		public:
			typedef basic_rawsocket<SOCK_STREAM, Family> rawsocket;
			typedef basic_sockaddress<Family> sockaddress;
			typedef CharT char_type;
			typedef typename std::basic_string<CharT, Traits> string_type;
			
			// constructor and destructor
			basic_socket() : rawsocket(), addr_() {}
			
			explicit basic_socket(const socket_type& s, const sockaddress& addr) :
				rawsocket(s), addr_(addr) {}
			
			explicit basic_socket(const char_type* host, int port) :
				rawsocket(), addr_() { this->connect(host, port); }
			
			explicit basic_socket(const string_type& host, int port) :
				rawsocket(), addr_() { this->connect(host, port); }
			
			virtual ~basic_socket() {}
			
			// operator
			basic_socket& connect(const char_type* host, int port) {
				if (this->open() < 0) throw clx::socket_error("socket");
				addr_.assign(host, port);
				if (::connect(this->socket(), addr_.base_data(), addr_.size()) == -1) {
					throw clx::socket_error("connect");
				}
				return *this;
			}
			
			basic_socket& connect(const string_type& host, int port) {
				return this->connect(host.c_str(), port);
			}
			
			int send(const char_type* dat, int n) {
				return ::send(this->socket(), dat, n, 0);
			}
			
			int recv(char_type* buf, int n) {
				return ::recv(this->socket(), buf, n, 0);
			}
			
			const sockaddress& address() const { return addr_; }
			
		private:
			sockaddress addr_;
		};
		
		/* ----------------------------------------------------------------- */
		//  basic_acceptor
		/* ----------------------------------------------------------------- */
		template <int Family>
		class basic_acceptor : public clx::basic_rawsocket<SOCK_STREAM, Family> {
		public:
			typedef basic_rawsocket<SOCK_STREAM, Family> rawsocket;
			typedef basic_sockaddress<Family> sockaddress;
			
			static const int nlisten = 100;
			static const int nretry = 10;
			
			// constructor and destructor
			basic_acceptor() : rawsocket() {}
			
			explicit basic_acceptor(int port, int n = nlisten) :
				rawsocket() { this->open(port, n); }
			
			virtual ~basic_acceptor() {}
			
			// operator
			basic_acceptor& open(int port, int n = nlisten) {
				if (rawsocket::open() < 0) throw socket_error("socket");
				addr_.assign(port);
				if (::bind(this->socket(), addr_.base_data(), addr_.size()) == -1) {
					throw socket_error("bind");
				}
				if (::listen(this->socket(), n) == -1) throw socket_error("listen");
				return *this;
			}
			
			basic_socket<Family> accept(int retry = nretry) {
				socket_type s = -1;
				sockaddress addr;
				typename sockaddress::inet_type tmp;
				socklen_type len = sizeof(tmp);
				
				for (int i = 0; i < retry; i++) {
					s = ::accept(this->socket(), (struct sockaddr*)&tmp, &len);
					if (s == -1) ::sleep(1);
					else break;
				}
				addr = tmp;
				return basic_socket<Family>(s, addr);
			}
		
		private:
			sockaddress addr_;
		};
		
		typedef basic_socket<AF_INET> socket;
		typedef basic_acceptor<AF_INET> acceptor;
		typedef basic_sockstream<socket> sockstream;
	};
}

#endif // CLX_TCP_H
